@@ -1,8 +1,6 @@
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
-#ifdef HAVE_SQLITE3_ERRSTR
+#if HAVE_SQLITE3_ERRSTR
 
 int dummy;
 
