@@ -1,7 +1,7 @@
-/*	$Id: mdoc.c,v 1.223 2014/08/06 15:09:05 schwarze Exp $ */
+/*	$Id: mdoc.c,v 1.238 2015/02/12 13:00:52 schwarze Exp $ */
 /*
  * Copyright (c) 2008, 2009, 2010, 2011 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2010, 2012, 2013, 2014 Ingo Schwarze <schwarze@openbsd.org>
+ * Copyright (c) 2010, 2012-2015 Ingo Schwarze <schwarze@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -15,9 +15,7 @@
  * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include <sys/types.h>
 
@@ -91,11 +89,7 @@ static	void		  mdoc_free1(struct mdoc *);
 static	void		  mdoc_alloc1(struct mdoc *);
 static	struct mdoc_node *node_alloc(struct mdoc *, int, int,
 				enum mdoct, enum mdoc_type);
-static	int		  node_append(struct mdoc *,
-				struct mdoc_node *);
-#if 0
-static	int		  mdoc_preptext(struct mdoc *, int, char *, int);
-#endif
+static	void		  node_append(struct mdoc *, struct mdoc_node *);
 static	int		  mdoc_ptext(struct mdoc *, int, char *, int);
 static	int		  mdoc_pmacro(struct mdoc *, int, char *, int);
 
@@ -196,41 +190,35 @@ mdoc_alloc(struct roff *roff, struct mparse *parse,
 	return(p);
 }
 
-int
+void
 mdoc_endparse(struct mdoc *mdoc)
 {
 
-	return(mdoc_macroend(mdoc));
+	mdoc_macroend(mdoc);
 }
 
-int
+void
 mdoc_addeqn(struct mdoc *mdoc, const struct eqn *ep)
 {
 	struct mdoc_node *n;
 
 	n = node_alloc(mdoc, ep->ln, ep->pos, MDOC_MAX, MDOC_EQN);
 	n->eqn = ep;
-
-	if ( ! node_append(mdoc, n))
-		return(0);
-
+	if (ep->ln > mdoc->last->line)
+		n->flags |= MDOC_LINE;
+	node_append(mdoc, n);
 	mdoc->next = MDOC_NEXT_SIBLING;
-	return(1);
 }
 
-int
+void
 mdoc_addspan(struct mdoc *mdoc, const struct tbl_span *sp)
 {
 	struct mdoc_node *n;
 
 	n = node_alloc(mdoc, sp->line, 0, MDOC_MAX, MDOC_TBL);
 	n->span = sp;
-
-	if ( ! node_append(mdoc, n))
-		return(0);
-
+	node_append(mdoc, n);
 	mdoc->next = MDOC_NEXT_SIBLING;
-	return(1);
 }
 
 /*
@@ -241,7 +229,8 @@ int
 mdoc_parseln(struct mdoc *mdoc, int ln, char *buf, int offs)
 {
 
-	mdoc->flags |= MDOC_NEWLINE;
+	if (mdoc->last->type != MDOC_EQN || ln > mdoc->last->line)
+		mdoc->flags |= MDOC_NEWLINE;
 
 	/*
 	 * Let the roff nS register switch SYNOPSIS mode early,
@@ -259,7 +248,7 @@ mdoc_parseln(struct mdoc *mdoc, int ln, char *buf, int offs)
 	    mdoc_ptext(mdoc, ln, buf, offs));
 }
 
-int
+void
 mdoc_macro(MACRO_PROT_ARGS)
 {
 	assert(tok < MDOC_MAX);
@@ -269,7 +258,7 @@ mdoc_macro(MACRO_PROT_ARGS)
 			mandoc_vmsg(MANDOCERR_DT_LATE,
 			    mdoc->parse, line, ppos,
 			    "Dt %s", buf + *pos);
-			return(1);
+			return;
 		}
 	} else if ( ! (mdoc_macros[tok].flags & MDOC_PROLOGUE)) {
 		if (mdoc->meta.title == NULL) {
@@ -282,12 +271,11 @@ mdoc_macro(MACRO_PROT_ARGS)
 			mdoc->meta.vol = mandoc_strdup("LOCAL");
 		mdoc->flags |= MDOC_PBODY;
 	}
-
-	return((*mdoc_macros[tok].fp)(mdoc, tok, line, ppos, pos, buf));
+	(*mdoc_macros[tok].fp)(mdoc, tok, line, ppos, pos, buf);
 }
 
 
-static int
+static void
 node_append(struct mdoc *mdoc, struct mdoc_node *p)
 {
 
@@ -331,8 +319,7 @@ node_append(struct mdoc *mdoc, struct mdoc_node *p)
 		break;
 	}
 
-	if ( ! mdoc_valid_pre(mdoc, p))
-		return(0);
+	mdoc_valid_pre(mdoc, p);
 
 	switch (p->type) {
 	case MDOC_HEAD:
@@ -359,14 +346,11 @@ node_append(struct mdoc *mdoc, struct mdoc_node *p)
 	case MDOC_TBL:
 		/* FALLTHROUGH */
 	case MDOC_TEXT:
-		if ( ! mdoc_valid_post(mdoc))
-			return(0);
+		mdoc_valid_post(mdoc);
 		break;
 	default:
 		break;
 	}
-
-	return(1);
 }
 
 static struct mdoc_node *
@@ -379,7 +363,6 @@ node_alloc(struct mdoc *mdoc, int line, int pos,
 	p->sec = mdoc->lastsec;
 	p->line = line;
 	p->pos = pos;
-	p->lastline = line;
 	p->tok = tok;
 	p->type = type;
 
@@ -396,62 +379,58 @@ node_alloc(struct mdoc *mdoc, int line, int pos,
 	return(p);
 }
 
-int
+void
 mdoc_tail_alloc(struct mdoc *mdoc, int line, int pos, enum mdoct tok)
 {
 	struct mdoc_node *p;
 
 	p = node_alloc(mdoc, line, pos, tok, MDOC_TAIL);
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_CHILD;
-	return(1);
 }
 
-int
+struct mdoc_node *
 mdoc_head_alloc(struct mdoc *mdoc, int line, int pos, enum mdoct tok)
 {
 	struct mdoc_node *p;
 
 	assert(mdoc->first);
 	assert(mdoc->last);
-
 	p = node_alloc(mdoc, line, pos, tok, MDOC_HEAD);
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_CHILD;
-	return(1);
+	return(p);
 }
 
-int
+struct mdoc_node *
 mdoc_body_alloc(struct mdoc *mdoc, int line, int pos, enum mdoct tok)
 {
 	struct mdoc_node *p;
 
 	p = node_alloc(mdoc, line, pos, tok, MDOC_BODY);
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_CHILD;
-	return(1);
+	return(p);
 }
 
-int
+struct mdoc_node *
 mdoc_endbody_alloc(struct mdoc *mdoc, int line, int pos, enum mdoct tok,
 		struct mdoc_node *body, enum mdoc_endbody end)
 {
 	struct mdoc_node *p;
 
+	body->flags |= MDOC_ENDED;
+	body->parent->flags |= MDOC_ENDED;
 	p = node_alloc(mdoc, line, pos, tok, MDOC_BODY);
-	p->pending = body;
+	p->body = body;
 	p->norm = body->norm;
 	p->end = end;
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_SIBLING;
-	return(1);
+	return(p);
 }
 
-int
+struct mdoc_node *
 mdoc_block_alloc(struct mdoc *mdoc, int line, int pos,
 		enum mdoct tok, struct mdoc_arg *args)
 {
@@ -477,14 +456,12 @@ mdoc_block_alloc(struct mdoc *mdoc, int line, int pos,
 	default:
 		break;
 	}
-
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_CHILD;
-	return(1);
+	return(p);
 }
 
-int
+void
 mdoc_elem_alloc(struct mdoc *mdoc, int line, int pos,
 		enum mdoct tok, struct mdoc_arg *args)
 {
@@ -502,26 +479,19 @@ mdoc_elem_alloc(struct mdoc *mdoc, int line, int pos,
 	default:
 		break;
 	}
-
-	if ( ! node_append(mdoc, p))
-		return(0);
+	node_append(mdoc, p);
 	mdoc->next = MDOC_NEXT_CHILD;
-	return(1);
 }
 
-int
+void
 mdoc_word_alloc(struct mdoc *mdoc, int line, int pos, const char *p)
 {
 	struct mdoc_node *n;
 
 	n = node_alloc(mdoc, line, pos, MDOC_MAX, MDOC_TEXT);
 	n->string = roff_strdup(mdoc->roff, p);
-
-	if ( ! node_append(mdoc, n))
-		return(0);
-
+	node_append(mdoc, n);
 	mdoc->next = MDOC_NEXT_SIBLING;
-	return(1);
 }
 
 void
@@ -603,68 +573,14 @@ mdoc_node_delete(struct mdoc *mdoc, struct mdoc_node *p)
 	mdoc_node_free(p);
 }
 
-int
+void
 mdoc_node_relink(struct mdoc *mdoc, struct mdoc_node *p)
 {
 
 	mdoc_node_unlink(mdoc, p);
-	return(node_append(mdoc, p));
+	node_append(mdoc, p);
 }
 
-#if 0
-/*
- * Pre-treat a text line.
- * Text lines can consist of equations, which must be handled apart from
- * the regular text.
- * Thus, use this function to step through a line checking if it has any
- * equations embedded in it.
- * This must handle multiple equations AND equations that do not end at
- * the end-of-line, i.e., will re-enter in the next roff parse.
- */
-static int
-mdoc_preptext(struct mdoc *mdoc, int line, char *buf, int offs)
-{
-	char		*start, *end;
-	char		 delim;
-
-	while ('\0' != buf[offs]) {
-		/* Mark starting position if eqn is set. */
-		start = NULL;
-		if ('\0' != (delim = roff_eqndelim(mdoc->roff)))
-			if (NULL != (start = strchr(buf + offs, delim)))
-				*start++ = '\0';
-
-		/* Parse text as normal. */
-		if ( ! mdoc_ptext(mdoc, line, buf, offs))
-			return(0);
-
-		/* Continue only if an equation exists. */
-		if (NULL == start)
-			break;
-
-		/* Read past the end of the equation. */
-		offs += start - (buf + offs);
-		assert(start == &buf[offs]);
-		if (NULL != (end = strchr(buf + offs, delim))) {
-			*end++ = '\0';
-			while (' ' == *end)
-				end++;
-		}
-
-		/* Parse the equation itself. */
-		roff_openeqn(mdoc->roff, NULL, line, offs, buf);
-
-		/* Process a finished equation? */
-		if (roff_closeeqn(mdoc->roff))
-			if ( ! mdoc_addeqn(mdoc, roff_eqn(mdoc->roff)))
-				return(0);
-		offs += (end - (buf + offs));
-	}
-
-	return(1);
-}
-#endif
-
 /*
  * Parse free-form text, that is, a line that does not begin with the
  * control character.
@@ -685,11 +601,12 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 	 * process within its context in the normal way).
 	 */
 
-	if (MDOC_Bl == n->tok && MDOC_BODY == n->type &&
-	    LIST_column == n->norm->Bl.type) {
+	if (n->tok == MDOC_Bl && n->type == MDOC_BODY &&
+	    n->end == ENDBODY_NOT && n->norm->Bl.type == LIST_column) {
 		/* `Bl' is open without any children. */
 		mdoc->flags |= MDOC_FREECOL;
-		return(mdoc_macro(mdoc, MDOC_It, line, offs, &offs, buf));
+		mdoc_macro(mdoc, MDOC_It, line, offs, &offs, buf);
+		return(1);
 	}
 
 	if (MDOC_It == n->tok && MDOC_BLOCK == n->type &&
@@ -698,7 +615,8 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 	    LIST_column == n->parent->norm->Bl.type) {
 		/* `Bl' has block-level `It' children. */
 		mdoc->flags |= MDOC_FREECOL;
-		return(mdoc_macro(mdoc, MDOC_It, line, offs, &offs, buf));
+		mdoc_macro(mdoc, MDOC_It, line, offs, &offs, buf);
+		return(1);
 	}
 
 	/*
@@ -746,7 +664,7 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 		mandoc_msg(MANDOCERR_SPACE_EOL, mdoc->parse,
 		    line, (int)(ws-buf), NULL);
 
-	if ('\0' == buf[offs] && ! (MDOC_LITERAL & mdoc->flags)) {
+	if (buf[offs] == '\0' && ! (mdoc->flags & MDOC_LITERAL)) {
 		mandoc_msg(MANDOCERR_FI_BLANK, mdoc->parse,
 		    line, (int)(c - buf), NULL);
 
@@ -755,18 +673,15 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 		 * blank lines aren't allowed, but enough manuals assume this
 		 * behaviour that we want to work around it.
 		 */
-		if ( ! mdoc_elem_alloc(mdoc, line, offs, MDOC_sp, NULL))
-			return(0);
-
+		mdoc_elem_alloc(mdoc, line, offs, MDOC_sp, NULL);
 		mdoc->next = MDOC_NEXT_SIBLING;
-
-		return(mdoc_valid_post(mdoc));
+		mdoc_valid_post(mdoc);
+		return(1);
 	}
 
-	if ( ! mdoc_word_alloc(mdoc, line, offs, buf+offs))
-		return(0);
+	mdoc_word_alloc(mdoc, line, offs, buf+offs);
 
-	if (MDOC_LITERAL & mdoc->flags)
+	if (mdoc->flags & MDOC_LITERAL)
 		return(1);
 
 	/*
@@ -779,7 +694,6 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 
 	if (mandoc_eos(buf+offs, (size_t)(end-buf-offs)))
 		mdoc->last->flags |= MDOC_EOS;
-
 	return(1);
 }
 
@@ -790,46 +704,47 @@ mdoc_ptext(struct mdoc *mdoc, int line, char *buf, int offs)
 static int
 mdoc_pmacro(struct mdoc *mdoc, int ln, char *buf, int offs)
 {
+	struct mdoc_node *n;
+	const char	 *cp;
 	enum mdoct	  tok;
 	int		  i, sv;
 	char		  mac[5];
-	struct mdoc_node *n;
-
-	/* Empty post-control lines are ignored. */
-
-	if ('"' == buf[offs]) {
-		mandoc_msg(MANDOCERR_COMMENT_BAD, mdoc->parse,
-		    ln, offs, NULL);
-		return(1);
-	} else if ('\0' == buf[offs])
-		return(1);
 
 	sv = offs;
 
 	/*
 	 * Copy the first word into a nil-terminated buffer.
-	 * Stop copying when a tab, space, or eoln is encountered.
+	 * Stop when a space, tab, escape, or eoln is encountered.
 	 */
 
 	i = 0;
-	while (i < 4 && '\0' != buf[offs] && ' ' != buf[offs] &&
-	    '\t' != buf[offs])
+	while (i < 4 && strchr(" \t\\", buf[offs]) == NULL)
 		mac[i++] = buf[offs++];
 
 	mac[i] = '\0';
 
 	tok = (i > 1 && i < 4) ? mdoc_hash_find(mac) : MDOC_MAX;
 
-	if (MDOC_MAX == tok) {
+	if (tok == MDOC_MAX) {
 		mandoc_msg(MANDOCERR_MACRO, mdoc->parse,
 		    ln, sv, buf + sv - 1);
 		return(1);
 	}
 
-	/* Disregard the first trailing tab, if applicable. */
+	/* Skip a leading escape sequence or tab. */
 
-	if ('\t' == buf[offs])
+	switch (buf[offs]) {
+	case '\\':
+		cp = buf + offs + 1;
+		mandoc_escape(&cp, NULL, NULL);
+		offs = cp - buf;
+		break;
+	case '\t':
 		offs++;
+		break;
+	default:
+		break;
+	}
 
 	/* Jump to the next non-whitespace word. */
 
@@ -850,8 +765,10 @@ mdoc_pmacro(struct mdoc *mdoc, int ln, char *buf, int offs)
 	 * into macro processing.
 	 */
 
-	if (NULL == mdoc->last || MDOC_It == tok || MDOC_El == tok)
-		return(mdoc_macro(mdoc, tok, ln, sv, &offs, buf));
+	if (NULL == mdoc->last || MDOC_It == tok || MDOC_El == tok) {
+		mdoc_macro(mdoc, tok, ln, sv, &offs, buf);
+		return(1);
+	}
 
 	n = mdoc->last;
 	assert(mdoc->last);
@@ -861,10 +778,11 @@ mdoc_pmacro(struct mdoc *mdoc, int ln, char *buf, int offs)
 	 * context around the parsed macro.
 	 */
 
-	if (MDOC_Bl == n->tok && MDOC_BODY == n->type &&
-	    LIST_column == n->norm->Bl.type) {
+	if (n->tok == MDOC_Bl && n->type == MDOC_BODY &&
+	    n->end == ENDBODY_NOT && n->norm->Bl.type == LIST_column) {
 		mdoc->flags |= MDOC_FREECOL;
-		return(mdoc_macro(mdoc, MDOC_It, ln, sv, &sv, buf));
+		mdoc_macro(mdoc, MDOC_It, ln, sv, &sv, buf);
+		return(1);
 	}
 
 	/*
@@ -878,13 +796,13 @@ mdoc_pmacro(struct mdoc *mdoc, int ln, char *buf, int offs)
 	    MDOC_Bl == n->parent->tok &&
 	    LIST_column == n->parent->norm->Bl.type) {
 		mdoc->flags |= MDOC_FREECOL;
-		return(mdoc_macro(mdoc, MDOC_It, ln, sv, &sv, buf));
+		mdoc_macro(mdoc, MDOC_It, ln, sv, &sv, buf);
+		return(1);
 	}
 
 	/* Normal processing of a macro. */
 
-	if ( ! mdoc_macro(mdoc, tok, ln, sv, &offs, buf))
-		return(0);
+	mdoc_macro(mdoc, tok, ln, sv, &offs, buf);
 
 	/* In quick mode (for mandocdb), abort after the NAME section. */
 
